(* 
 For theory morphisms adapted file: 

    Title:      Pure/Proof/proofchecker.ML
    Author:     Stefan Berghofer, TU Muenchen

    Simple proof checker based only on the core inference rules
    of Isabelle/Pure.

Maps a proof term of a given thm by given mappings and derives
the new thm in the target theory.

*)


signature DERIVE_THM =
sig

  datatype AxmImg = Axm of Symtab.key
                  | Thm of Symtab.key

  val derive_thm : (Term.typ -> Term.typ) -> (Term.term -> Term.term) ->
                   Symtab.key Symtab.table -> AxmImg Symtab.table -> Symtab.key Symtab.table -> 
                   Context.theory -> Context.theory -> Proofterm.proof -> thm

  val derive_thmI : Context.theory -> Proofterm.proof -> thm

  exception DERIVE_THM of string

end

structure Derive_Thm : DERIVE_THM =
struct




exception DERIVE_THM of string

datatype AxmImg = Axm of Symtab.key
                | Thm of Symtab.key

fun prt_verb s = (if false then writeln s else ())

open Proofterm Aux;

(***** construct a theorem out of a proof term *****)

fun lookup_thm thy =
  let val tab = fold_rev Symtab.update (PureThy.all_thms_of thy) Symtab.empty
  in
    (fn s => case Symtab.lookup tab s of
       NONE => raise DERIVE_THM ("Unknown theorem " ^ quote s)
     | SOME thm => thm)
  end;


fun try_with_suffix thy get_thm mp nm =
 let val nm' = nm ^ "_1"
 in
   case (Symtab.lookup mp nm') of
    SOME i => SOME (get_thm thy i)    
  | NONE => NONE
 end




fun thm_of_proof map_type map_term defMap axmMap thmMap src trg prf =
  let fun get_prf_names p = Proofterm.fold_proof_terms
          (fold_aterms (fn Free (x, _) => Name.declare x | _ => I)) (K I) p Name.context
      val prf_names = get_prf_names prf
      val lookup = lookup_thm trg 

    fun thm_of_atom thm Ts =
      let
        val tvars = term_tvars (Thm.full_prop_of thm);
        val (fmap, thm') = Thm.varifyT' [] thm;
        val ctye = map (pairself (Thm.ctyp_of trg))
          (map TVar tvars @ map (fn ((_, S), ixn) => TVar (ixn, S)) fmap ~~ Ts)
      in
        Thm.instantiate (ctye, []) (forall_intr_vars (forall_intr_frees thm'))
      end


    fun thm_of vs Hs (prf as (PThm ((name, prf', prop', SOME Ts)))) =
    (let val img = (case (Symtab.lookup thmMap name) of
                     SOME i => Drule.implies_intr_hyps (lookup i)
                   | NONE => 
                      (case prf' of
                        (PAxm (name', _, _)) => (case (Symtab.lookup defMap name') of
                                                  SOME i => get_axiom trg i
                                                | NONE => (case (Symtab.lookup axmMap name') of
                                                            SOME (Axm i) => get_axiom trg i
                                                          | SOME (Thm i) => Drule.implies_intr_hyps (lookup i)
                                                          | NONE => get_axiom trg name'))
                      | _ => Drule.implies_intr_hyps (lookup name)
                             handle e => (let val _ = writeln ("!!! Theorem not found : " ^ name)
                                              val p = (Aux.strip_thm (Aux.freeze 
                                                      (Aux.reconstruct_prf' src prop' name prf')))
                                              val t = thm_of ([], get_prf_names p) [] p 
                                              val t' = Aux.beta_eta_convert t 
                                              val t'' = (Drule.standard o Thm.varifyT) t'
                                          in
                                           Drule.implies_intr_hyps t''
                                          end)))
         val TTs = map map_type Ts
      in
        thm_of_atom img TTs
      end)

      | thm_of _ _ (PAxm (name, _, SOME Ts)) =
        let val img = (case (Symtab.lookup axmMap name) of
                                       SOME (Axm i) => get_axiom trg i
                                     | SOME (Thm i) => Drule.implies_intr_hyps (lookup i) 
                                     | NONE => (case (Symtab.lookup defMap name) of
                                                  SOME i => get_axiom trg i 
                                                | NONE => get_axiom trg name)) 
            val TTs = map map_type Ts
        in
          thm_of_atom img TTs
        end

      | thm_of _ Hs (PBound i) = List.nth (Hs, i)
                                   
      | thm_of (vs, names) Hs (Abst (s, SOME T, prf)) =
        let val TT = map_type T
            val ([x], names') = Name.variants [s] names;
            val thm = thm_of ((x, TT) :: vs, names') Hs prf
        in
            Thm.forall_intr (Thm.cterm_of trg (Free (x, TT))) thm
        end

      | thm_of (vs, names) Hs (prf % SOME t) =
          let val tt' = map_term t
              val tt = Term.subst_bounds ((map Free vs), tt')
              val ct = Thm.cterm_of trg tt
              val thm = thm_of (vs, names) Hs prf

              fun alpha_thm vars Hyps proof cterm thm =
               let val prop = prop_of thm
                   val bvar_type = (case prop of
                                     Const ("all", Type ("fun", [Type ("fun", [qary, _]), _])) $ A => qary
                                   | _ =>  raise DERIVE_THM ("forall_elim: not quantified"))
                   val {T=tt_type, ...} = Thm.rep_cterm cterm

                   val athm = (case (Alpha.typ_aconv_map (Sign.tsig_of trg) (bvar_type, tt_type)) of
                                NONE => (let val _ = Output.error_msg ("forall_elim alpha-conversion failure: "^
                                        ("\nleft: " ^ ((Display.string_of_ctyp o (Thm.ctyp_of trg)) bvar_type) ^
                                         "\nright: "^ ((Display.string_of_ctyp o (Thm.ctyp_of trg)) tt_type)))
                                         in
                                           raise DERIVE_THM ("forall_elim: type mismatch")  
                                         end) 
                             | SOME mp => 
                              (let val proof' = Proofterm.map_proof_terms (Alpha.map_aconv_term mp)
                                                                          (Alpha.map_aconv_typ mp) proof
                               in
                                 thm_of vars Hyps proof'
                               end))
                in
                 athm
                end                              
               
          in 
           Thm.forall_elim ct thm
           handle THM (_, 0, _) => (let val _ = prt_verb ("Trying alpha for ... " ^ (string_of_thm thm))
                                        val athm = alpha_thm (vs, names) Hs prf ct thm
                                        val _ = prt_verb ("completed. ")
                                    in
                                      Thm.forall_elim ct athm
                                    end)
          end

      | thm_of (vs, names) Hs (AbsP (s, SOME t, prf)) =
          let val tt' = map_term t
              val tt = Term.subst_bounds ((map Free vs), tt')
              val ct = Thm.cterm_of trg tt
              val thm = thm_of (vs, names) (Thm.assume ct :: Hs) prf

          in
            Thm.implies_intr ct thm
          end

      | thm_of vars HsT (prf %% prf') =
          let val thmAB = Aux.beta_eta_convert (thm_of vars HsT prf)
              val thmA = Aux.beta_eta_convert (thm_of vars HsT prf')
          in
            Thm.implies_elim thmAB thmA
          end
  
      | thm_of _ _ (Hyp t) = 
         let val tt = map_term t
         in
          Thm.assume (Thm.cterm_of trg tt)
         end

      | thm_of _ _ _ = raise DERIVE_THM "partial proof term"

  in 
   Aux.beta_eta_convert (thm_of ([], prf_names) [] prf) 
  end


fun derive_thm map_type map_term defMap axmMap thmMap src trg prf =
 let val thm = thm_of_proof map_type map_term defMap axmMap thmMap src trg (Aux.freeze prf)
 in
  (Drule.standard o Thm.varifyT) thm
 end



(* derive without mapping *)

fun thm_of_proofI thy prf =
  let fun get_prf_names p = Proofterm.fold_proof_terms
          (fold_aterms (fn Free (x, _) => Name.declare x | _ => I)) (K I) p Name.context
      val prf_names = get_prf_names prf
      val lookup = lookup_thm thy 

      fun thm_of_atom thm Ts =
       let val tvars = term_tvars (Thm.full_prop_of thm);
           val (fmap, thm') = Thm.varifyT' [] thm;
           val ctye = map (pairself (Thm.ctyp_of thy))
          (map TVar tvars @ map (fn ((_, S), ixn) => TVar (ixn, S)) fmap ~~ Ts)
       in
         Thm.instantiate (ctye, []) (forall_intr_vars (forall_intr_frees thm'))
       end


    fun thm_of vs Hs (prf as (PThm ((name, prf', prop', SOME Ts)))) =
    (let val img = (case prf' of
                        (PAxm (name', _, _)) => get_axiom thy name'
                      | _ => Drule.implies_intr_hyps (lookup name)
                             handle e => (let val _ = Aux.warning ("!!! Theorem not found : " ^ name)
                                              val p = (Aux.strip_thm (Aux.freeze 
                                                      (Aux.reconstruct_prf' thy prop' name prf')))
                                              val t = thm_of ([], get_prf_names p) [] p 
                                              val t' = Aux.beta_eta_convert t 
                                              val t'' = (Drule.standard o Thm.varifyT) t'
                                          in
                                           Drule.implies_intr_hyps t''
                                          end))
      in
        thm_of_atom img Ts
      end)

      | thm_of _ _ (PAxm (name, _, SOME Ts)) =
        let val img = get_axiom thy name 
        in
          thm_of_atom img Ts
        end

      | thm_of _ Hs (PBound i) = List.nth (Hs, i)
                                   
      | thm_of (vs, names) Hs (Abst (s, SOME T, prf)) =
        let val ([x], names') = Name.variants [s] names;
            val thm = thm_of ((x, T) :: vs, names') Hs prf
        in
            Thm.forall_intr (Thm.cterm_of thy (Free (x, T))) thm
        end

      | thm_of (vs, names) Hs (prf % SOME t) =
          let val tt = Term.subst_bounds ((map Free vs), t)
              val ct = Thm.cterm_of thy tt
              val thm = thm_of (vs, names) Hs prf

              fun alpha_thm vars Hyps proof cterm thm =
               let val _ = Aux.writeln "Trying alpha conversion"
                   val prop = prop_of thm
                   val bvar_type = (case prop of
                                     Const ("all", Type ("fun", [Type ("fun", [qary, _]), _])) $ A => qary
                                   | _ =>  raise DERIVE_THM ("forall_elim: not quantified"))
                   val {T=tt_type, ...} = Thm.rep_cterm cterm

                   val athm = (case (Alpha.typ_aconv_map (Sign.tsig_of thy) (bvar_type, tt_type)) of
                                NONE => (let val _ = Output.error_msg ("forall_elim alpha-conversion failure: "^
                                        ("\nleft: " ^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) bvar_type) ^
                                         "\nright: "^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) tt_type)))
                                         in
                                           raise DERIVE_THM ("forall_elim: type mismatch")  
                                         end) 
                             | SOME mp => 
                              (let val proof' = Proofterm.map_proof_terms (Alpha.map_aconv_term mp)
                                                                          (Alpha.map_aconv_typ mp) proof
                                   val vars' = (map (fn (v, t) => (v, Alpha.map_aconv_typ mp t)) (fst vars),
                                                snd vars)
                               in
                                 thm_of vars Hyps proof'
                               end))
                in
                 athm
                end                              
               
          in 
           Thm.forall_elim ct thm
           handle THM (_, 0, _) => (let val athm = alpha_thm (vs, names) Hs prf ct thm
                                    in
                                      Thm.forall_elim ct athm
                                    end)
          end

      | thm_of (vs, names) Hs (AbsP (s, SOME t, prf)) =
          let val tt = Term.subst_bounds ((map Free vs), t)
              val ct = Thm.cterm_of thy tt
              val thm = thm_of (vs, names) (Thm.assume ct :: Hs) prf
          in
            Thm.implies_intr ct thm
          end

      | thm_of vars HsT (prf %% prf') =
          let val thmAB = Aux.beta_eta_convert (thm_of vars HsT prf)
              val thmA = Aux.beta_eta_convert (thm_of vars HsT prf')
          in
            Thm.implies_elim thmAB thmA
          end
  
      | thm_of _ _ (Hyp t) = Thm.assume (Thm.cterm_of thy t)

      | thm_of _ _ _ = raise DERIVE_THM "partial proof term"

  in 
   Aux.beta_eta_convert (thm_of ([], prf_names) [] prf) 
  end


fun derive_thmI thy prf =
 let val thm = thm_of_proofI thy (Aux.freeze prf) 
 in
  (Drule.standard o Thm.varifyT) thm
 end

end
